# pylint: disable=duplicate-code

import platform
import subprocess as sub

import numpy as np
import pandas as pd
import seaborn as sns


# ============================================
#                   clear
# ============================================
def clear() -> None:
    try:
        sub.run(
            ["cls" if "windows" == platform.system().lower() else "clear"], check=True
        )
    # For git bash
    except FileNotFoundError:
        sub.run(["clear"], check=True)


# ============================================
#                    plot
# ============================================
def plot(desired, measured, times, label, figName):
    nValues = len(times)
    t = np.concatenate((times, times))
    dataType = ["Desired"] * nValues + ["Measured"] * nValues
    data = np.concatenate((desired, measured))

    data = {
        "Time (ms)": t,
        "Type": dataType,
        label: data,
    }

    df = pd.DataFrame(data)

    plt = sns.relplot(data=df, x="Time (ms)", y=label, hue="Type", kind="line")

    plt.figure.savefig(figName)



# pylint: disable=duplicate-code

import argparse
from time import sleep

from utils import plot  # pylint: disable=import-error

from flexsea.device import Device


# ============================================
#                     main
# ============================================
# pylint: disable=too-many-locals
def main(
    port: str,
    cLibVersion: str,
    libFile: str,
    offset: int,
    gains: dict,
    runTime: int,
    transitionTime: int,
    commandDelay: int,
    freq: int,
):
    device = Device(port=port, cLibVersion=cLibVersion, libFile=libFile)
    device.open()
    device.start_streaming(freq)

    data = device.read()

    pos0 = data["mot_ang"]
    positions = [pos0, pos0 + offset]

    device.set_gains(**gains)

    positionIndex = 0

    # For plotting
    measuredPosition = []
    desiredPosition = []
    deviceTime = []

    nLoops = int(runTime / commandDelay)
    transitionSteps = int(transitionTime / commandDelay)

    # Oscillate between the two positions nLoops times
    for i in range(nLoops):
        data = device.read()

        # Save the data for plotting
        measuredPosition.append(data["mot_ang"])
        desiredPosition.append(positions[positionIndex])
        deviceTime.append(data["state_time"])

        # Every transitionTime seconds, we go to the next position
        if i % transitionSteps == 0:
            positionIndex = (positionIndex + 1) % len(positions)
            device.command_motor_position(int(positions[positionIndex]))

        sleep(commandDelay)

    device.close()
    plot(
        desiredPosition,
        measuredPosition,
        deviceTime,
        "Motor Position ('ticks')",
        "two_position.png",
    )


# ============================================
#                   Run Main
# ============================================
if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=str,
        default="",
        help="Serial port device is connected to.",
    )
    parser.add_argument(
        "-c",
        "--clibversion",
        dest="cLibVersion",
        type=str,
        default="",
        help="Semantic version string of the pre-compiled C library.",
    )
    parser.add_argument(
        "-l",
        "--libfile",
        dest="libFile",
        type=str,
        default="",
        help="Path to the pre-compiled C library to use.",
    )
    parser.add_argument(
        "-o",
        "--offset",
        dest="offset",
        type=int,
        default=1000,
        help="Number of motor ticks to move from initial position.",
    )
    parser.add_argument(
        "-t",
        "--run-time",
        dest="runTime",
        type=int,
        default=10,
        help="Duration of test in seconds.",
    )
    parser.add_argument(
        "--transition-time",
        dest="transitionTime",
        type=int,
        default=2,
        help="Number of seconds between position changes.",
    )
    parser.add_argument(
        "--command-delay",
        dest="commandDelay",
        type=float,
        default=0.1,
        help="Number of seconds to sleep after issuing a motor command.",
    )
    parser.add_argument(
        "-f",
        "--frequency",
        dest="freq",
        type=int,
        default=100,
        help="Frequency (Hz) at which device will stream data.",
    )
    parser.add_argument(
        "--kp",
        dest="kp",
        type=int,
        default=50,
        help="Proportional gain.",
    )
    parser.add_argument(
        "--ki",
        dest="ki",
        type=int,
        default=0,
        help="Integral gain.",
    )
    parser.add_argument(
        "--kd",
        dest="kd",
        type=int,
        default=0,
        help="Differential gain.",
    )
    parser.add_argument(
        "--k",
        dest="k",
        type=int,
        default=0,
        help="Stiffness gain.",
    )
    parser.add_argument(
        "--b",
        dest="b",
        type=int,
        default=0,
        help="Damping gain.",
    )
    parser.add_argument(
        "--ff",
        dest="ff",
        type=int,
        default=0,
        help="Feed-forward gain.",
    )

    args = parser.parse_args()

    gains_dict = {
        "kp": args.kp,
        "ki": args.ki,
        "kd": args.kd,
        "k": args.k,
        "b": args.b,
        "ff": args.ff,
    }

    main(
        args.port,
        args.cLibVersion,
        args.libFile,
        args.offset,
        gains_dict,
        args.runTime,
        args.transitionTime,
        args.commandDelay,
        args.freq,
    )

    print("Done.")



# pylint: disable=duplicate-code

import argparse
from time import sleep

import numpy as np
from utils import clear  # pylint: disable=import-error

from flexsea.device import Device


# ============================================
#                    main
# ============================================
def main(
    port: str,
    cLibVersion: str,
    libFile: str,
    freq: int,
    gains: dict,
    minCurrent: int,
    maxCurrent: int,
    nCurrents: int,
    commandDelay: float,
    holdTime: int,
):
    device = Device(port=port, cLibVersion=cLibVersion, libFile=libFile)
    device.open()
    device.start_streaming(freq)

    device.set_gains(**gains)

    currents = np.linspace(minCurrent, maxCurrent, nCurrents)

    for current in currents:
        device.command_motor_current(int(current))
        sleep(commandDelay)
        device.print()
        clear()

    sleep(holdTime)

    try:
        for current in currents[-1::-1]:
            device.command_motor_current(int(current))
            sleep(commandDelay)
            device.print()
            clear()
    except KeyboardInterrupt:
        device.stop_motor()
        sleep(commandDelay)

    device.close()


# ============================================
#                  Run Main
# ============================================
if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=str,
        default="",
        help="Serial port device is connected to.",
    )
    parser.add_argument(
        "-c",
        "--clibversion",
        dest="cLibVersion",
        type=str,
        default="",
        help="Semantic version string of the pre-compiled C library.",
    )
    parser.add_argument(
        "-l",
        "--libfile",
        dest="libFile",
        type=str,
        default="",
        help="Path to the pre-compiled C library to use.",
    )
    parser.add_argument(
        "-f",
        "--frequency",
        dest="freq",
        type=int,
        default=100,
        help="Frequency (Hz) at which device will stream data.",
    )
    parser.add_argument(
        "--kp",
        dest="kp",
        type=int,
        default=40,
        help="Proportional gain.",
    )
    parser.add_argument(
        "--ki",
        dest="ki",
        type=int,
        default=400,
        help="Integral gain.",
    )
    parser.add_argument(
        "--kd",
        dest="kd",
        type=int,
        default=0,
        help="Differential gain.",
    )
    parser.add_argument(
        "--k",
        dest="k",
        type=int,
        default=0,
        help="Stiffness gain.",
    )
    parser.add_argument(
        "--b",
        dest="b",
        type=int,
        default=0,
        help="Damping gain.",
    )
    parser.add_argument(
        "--ff",
        dest="ff",
        type=int,
        default=0,
        help="Feed-forward gain.",
    )
    parser.add_argument(
        "--max-current",
        dest="maxCurrent",
        type=int,
        default=1000,
        help="Maximum current to ramp to, in milliamps.",
    )
    parser.add_argument(
        "--min-current",
        dest="minCurrent",
        type=int,
        default=0,
        help="Minium current to start at, in milliamps.",
    )
    parser.add_argument(
        "-n",
        "--n-currents",
        dest="nCurrents",
        type=int,
        default=50,
        help="The number of currents to ramp over.",
    )
    parser.add_argument(
        "--command-delay",
        dest="commandDelay",
        type=float,
        default=0.1,
        help="Number of seconds to sleep after issuing a motor command.",
    )
    parser.add_argument(
        "--hold-time",
        dest="holdTime",
        type=int,
        default=3,
        help="The time (seconds) to hold peak current.",
    )

    args = parser.parse_args()

    gains_dict = {
        "kp": args.kp,
        "ki": args.ki,
        "kd": args.kd,
        "k": args.k,
        "b": args.b,
        "ff": args.ff,
    }

    main(
        args.port,
        args.cLibVersion,
        args.libFile,
        args.freq,
        gains_dict,
        args.minCurrent,
        args.maxCurrent,
        args.nCurrents,
        args.commandDelay,
        args.holdTime,
    )

    print("Done.")



"""
///
/// @file MD20State.py
///
/// @brief AUTOGENERATED FILE! ONLY EDIT IF YOU ARE A MACHINE!
///
/// @core c4674f111fc72bad86b69ef5e7f862f75f59d86f
///
/// @build 6a54779ef7f5d112e58f4877e4667f33623592a3
///
/// @date 2023-02-23 14:31:03
///
/// @author Dephy, Inc.
"""
# pylint: disable=duplicate-code

from ctypes import Structure, c_int


class MD20State(Structure):
    _pack_ = 1
    _fields_ = [
        ("device_family", c_int),
        ("variable_label", c_int),
        ("include files:", c_int),
        ("state_time", c_int),
        ("accelx", c_int),
        ("accely", c_int),
        ("accelz", c_int),
        ("gyrox", c_int),
        ("gyroy", c_int),
        ("gyroz", c_int),
        ("mot_ang", c_int),
        ("mot_vel", c_int),
        ("mot_acc", c_int),
        ("mot_cur", c_int),
        ("mot_volt", c_int),
        ("batt_volt", c_int),
        ("batt_curr", c_int),
        ("temperature", c_int),
        ("status_mn", c_int),
        ("status_ex", c_int),
        ("status_re", c_int),
        ("genvar_0", c_int),
        ("genvar_1", c_int),
        ("genvar_2", c_int),
        ("genvar_3", c_int),
        ("genvar_4", c_int),
        ("genvar_5", c_int),
        ("genvar_6", c_int),
        ("genvar_7", c_int),
        ("genvar_8", c_int),
        ("genvar_9", c_int),
        ("genvar_10", c_int),
        ("genvar_11", c_int),
        ("genvar_12", c_int),
        ("genvar_13", c_int),
        ("genvar_14", c_int),
        ("ank_ang", c_int),
        ("ank_vel", c_int),
        ("shank_ang", c_int),
        ("shank_vel", c_int),
        ("global_shank_ang", c_int),
        ("ank_pos_x", c_int),
        ("ank_pos_y", c_int),
        ("ank_pos_z", c_int),
        ("ank_linear_vel_x", c_int),
        ("ank_linear_vel_y", c_int),
        ("ank_linear_vel_z", c_int),
        ("mot_from_ank", c_int),
        ("ank_from_mot", c_int),
        ("trans_ratio", c_int),
        ("ank_torque", c_int),
        ("exo_step_power", c_int),
        ("step_count", c_int),
        ("step_time", c_int),
        ("gait_state", c_int),
        ("intermediate_state", c_int),
        ("movement", c_int),
        ("toe_whip", c_int),
        ("SystemTime", c_int),
    ]



# pylint: disable=duplicate-code

import argparse
from time import sleep

import numpy as np
from utils import plot  # pylint: disable=import-error

from flexsea.device import Device


# ============================================
#                 send_command
# ============================================
def send_command(
    cmd: dict, cmdType: str, gains: dict, setGains: bool, plotData: dict, device: Device
) -> dict:
    deviceData = device.read()

    if setGains:
        device.set_gains(**gains)

    setPoint = cmd["cur"] if cmdType == "current" else cmd["pos"]

    if cmdType == "current":
        device.command_motor_current(int(setPoint))
    else:
        device.command_motor_position(int(setPoint))

    plotData["stateTime"].append(deviceData["state_time"])
    plotData["desiredPosition"].append(cmd["pos"])
    plotData["measuredPosition"].append(deviceData["mot_ang"])
    plotData["desiredCurrent"].append(cmd["cur"])
    plotData["measuredCurrent"].append(deviceData["mot_cur"])

    return plotData


# ============================================
#                  get_gains
# ============================================
def get_gains() -> dict:
    positionGains = {
        "kp": 100,
        "ki": 10,
        "kd": 0,
        "k": 0,
        "b": 0,
        "ff": 0,
    }

    currentGains = {
        "kp": 40,
        "ki": 400,
        "kd": 0,
        "k": 0,
        "b": 0,
        "ff": 128,
    }

    gains = {
        "current": currentGains,
        "position": positionGains,
    }

    return gains


# ============================================
#                 get_samples
# ============================================
def get_samples(frequency: int) -> dict:
    # Position samples
    positionAmplitude = 10000
    positionFreq = 1

    nSamples = int(frequency / positionFreq)
    x = np.linspace(-np.pi, np.pi, nSamples)
    positionSamples = positionAmplitude * np.sin(x)

    # Current samples
    currentAmplitude = 1500
    currentFreq = 5

    nSamples = int(frequency / currentFreq)
    x = np.linspace(-np.pi, np.pi, nSamples)
    currentSamples = currentAmplitude * np.sin(x)

    # Current samples: line
    value = 0
    length = 0.5

    nSamples = np.int32(length * frequency)
    currentSamplesLine = np.array([value for _ in range(nSamples)])

    # Samples
    samples = {
        "position": positionSamples,
        "currentSine": currentSamples,
        "currentLine": currentSamplesLine,
    }

    return samples


# ============================================
#                    main
# ============================================
def main(
    port: str, cLibVersion: str, libFile: str, freq: int, nLoops: int
):  # pylint: disable=too-many-locals
    delay = 1.0 / freq
    # Factor to determine current on "way back" to starting point
    currentAsymmetricG = 1.15

    device = Device(port=port, cLibVersion=cLibVersion, libFile=libFile)
    device.open()
    device.start_streaming(freq)

    # Initialize the gains to be used for each controller
    gains = get_gains()

    # Get the samples for each controller
    samples = get_samples(freq)

    # Initial device position
    pos0 = device.read()["mot_ang"]

    # Plotting arrays
    data = {
        "stateTime": [],
        "desiredPosition": [],
        "measuredPosition": [],
        "desiredCurrent": [],
        "measuredCurrent": [],
    }

    for i in range(nLoops):
        # Step 0
        sleep(delay)
        if i:
            command = {"cur": 0, "pos": device.read()["mot_ang"]}
        else:
            command = {"cur": 0, "pos": pos0}
        data = send_command(command, "position", gains["position"], True, data, device)

        # Step 1
        if i:
            for sample in np.linspace(data["mot_ang"], pos0, 360):
                command = {"cur": 0, "pos": sample}
                sleep(delay)
                data = send_command(command, "position", {}, False, data, device)

        # Step 2
        for sample in samples["position"]:
            command = {"cur": 0, "pos": sample + pos0}
            sleep(delay)
            data = send_command(command, "position", {}, False, data, device)

        # Step 3
        command = {"cur": 0, "pos": pos0}
        data = send_command(command, "current", gains["current"], True, data, device)
        sleep(delay)

        # Step 4
        for sample in samples["currentSine"].astype(np.int64):
            sleep(delay)
            if sample > 0:
                sample *= currentAsymmetricG
            command = {"cur": sample, "pos": pos0}
            sleep(delay)
            data = send_command(command, "current", {}, False, data, device)

        # Step 5
        for sample in samples["currentLine"]:
            command = {"cur": sample, "pos": pos0}
            sleep(delay)
            data = send_command(command, "current", {}, False, data, device)

    device.close()

    # Plot data
    times = [data["stateTime"], data["stateTime"]]
    desired = [data["desiredCurrent"], data["desiredPosition"]]
    measured = [data["measuredCurrent"], data["measuredPosition"]]
    labels = ["current", "position"]

    for t, des, meas, label in zip(times, desired, measured, labels):
        plot(des, meas, t, label, f"{label}_high_stress.png")


# ============================================
#                  Run Main
# ============================================
if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=str,
        default="",
        help="Serial port device is connected to.",
    )
    parser.add_argument(
        "-c",
        "--clibversion",
        dest="cLibVersion",
        type=str,
        default="",
        help="Semantic version string of the pre-compiled C library.",
    )
    parser.add_argument(
        "-l",
        "--libfile",
        dest="libFile",
        type=str,
        default="",
        help="Path to the pre-compiled C library to use.",
    )
    parser.add_argument(
        "-f",
        "--frequency",
        dest="freq",
        type=int,
        default=100,
        help="Frequency (Hz) at which device will stream data.",
    )
    parser.add_argument(
        "--n-loops",
        dest="nLoops",
        type=int,
        default=1,
        help="Feed-forward gain.",
    )

    args = parser.parse_args()

    main(args.port, args.cLibVersion, args.libFile, args.freq, args.nLoops)

    print("Done.")
